"""
Unit tests
"""
import pytest

import pepotron


@pytest.mark.parametrize(
    "search, expected_url",
    [
        ("8", "https://peps.python.org/pep-0008/"),
        ("12", "https://peps.python.org/pep-0012/"),
        ("2.7", "https://peps.python.org/pep-0373/"),
        (None, "https://peps.python.org"),
        ("dead batteries", "https://peps.python.org/pep-0594/"),
    ],
)
def test_url(search: str, expected_url: str) -> None:
    # Act
    pep_url = pepotron.url(search)
    # Assert
    assert pep_url == expected_url


@pytest.mark.parametrize(
    "search, base_url, expected_url",
    [
        (
            "8",
            "https://hugovk.github.io/peps",
            "https://hugovk.github.io/peps/pep-0008/",
        ),
        (
            "3.11",
            "https://hugovk.github.io/peps",
            "https://hugovk.github.io/peps/pep-0664/",
        ),
        (
            None,
            "https://hugovk.github.io/peps",
            "https://hugovk.github.io/peps",
        ),
        (
            "dead batteries",
            "https://hugovk.github.io/peps",
            "https://hugovk.github.io/peps/pep-0594/",
        ),
    ],
)
def test_url_base_url(search: str, base_url: str, expected_url: str) -> None:
    # Act
    pep_url = pepotron.url(search, base_url)
    # Assert
    assert pep_url == expected_url


@pytest.mark.parametrize(
    "search, expected_url",
    [
        ("594", "https://pep-previews--2440.org.readthedocs.build/pep-0594/"),
        (None, "https://pep-previews--2440.org.readthedocs.build"),
    ],
)
def test_url_pr(search, expected_url) -> None:
    # Arrange
    pr = 2440
    # Act
    pep_url = pepotron.url(search, pr=pr)
    # Assert
    assert pep_url == expected_url


def test_pep() -> None:
    pepotron.pep("8", dry_run=True)



"""
Unit tests for _cache
"""
import tempfile
from pathlib import Path

from freezegun import freeze_time

from pepotron import _cache


class TestCache:
    def setup_method(self):
        # Choose a new cache dir that doesn't exist
        self.original_cache_dir = _cache.CACHE_DIR
        self.temp_dir = tempfile.TemporaryDirectory()
        _cache.CACHE_DIR = Path(self.temp_dir.name) / "pepotron"

    def teardown_method(self):
        # Reset original
        _cache.CACHE_DIR = self.original_cache_dir

    @freeze_time("2018-12-26")
    def test__cache_filename(self):
        # Arrange
        url = "https://endoflife.date/api/python.json"

        # Act
        out = _cache.filename(url)

        # Assert
        # TODO rename
        assert str(out).endswith("2018-12-26-https-endoflife-date-api-python-json.json")

    def test__load_cache_not_exist(self):
        # Arrange
        filename = Path("file-does-not-exist")

        # Act
        data = _cache.load(filename)

        # Assert
        assert data == {}

    def test__load_cache_bad_data(self):
        # Arrange
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(b"Invalid JSON!")

        # Act
        data = _cache.load(Path(f.name))

        # Assert
        assert data == {}

    def test_cache_round_trip(self):
        # Arrange
        filename = _cache.CACHE_DIR / "test_cache_round_trip.json"
        data = {"1": "test data"}

        # Act
        _cache.save(filename, data)
        new_data = _cache.load(filename)

        # Tidy up
        filename.unlink()

        # Assert
        assert new_data == data

    @freeze_time("2021-10-25")
    def test__clear_cache_all(self):
        # Arrange
        # Create old cache file
        cache_file_old = _cache.CACHE_DIR / "2021-10-24-old-cache-file.json"
        cache_file_new = _cache.CACHE_DIR / "2021-10-25-new-cache-file.json"
        _cache.save(cache_file_old, data={})
        _cache.save(cache_file_new, data={})
        assert cache_file_new.exists()
        assert cache_file_old.exists()

        # Act
        _cache.clear(clear_all=True)

        # Assert
        assert not cache_file_old.exists()
        assert not cache_file_new.exists()

    @freeze_time("2021-10-25")
    def test__clear_cache_old(self):
        # Arrange
        # Create old cache file
        cache_file_old = _cache.CACHE_DIR / "2021-10-24-old-cache-file.json"
        cache_file_new = _cache.CACHE_DIR / "2021-10-25-new-cache-file.json"
        _cache.save(cache_file_old, data={})
        _cache.save(cache_file_new, data={})
        assert cache_file_new.exists()
        assert cache_file_old.exists()

        # Act
        _cache.clear()

        # Assert
        assert not cache_file_old.exists()
        assert cache_file_new.exists()



"""
CLI to open PEPs in your browser
"""
from __future__ import annotations

import logging
import warnings
from pathlib import Path

from pepotron import _cache

try:
    # Python 3.8+
    import importlib.metadata as importlib_metadata
except ImportError:
    # Python 3.7 and lower
    import importlib_metadata  # type: ignore

__version__ = importlib_metadata.version(__name__)

BASE_URL = "https://peps.python.org"
JSON_PATH = "/api/peps.json"
USER_AGENT = f"pepotron/{__version__}"

VERSION_TO_PEP = {
    "1.6": 160,
    "2.0": 200,
    "2.1": 226,
    "2.2": 251,
    "2.3": 283,
    "2.4": 320,
    "2.5": 356,
    "2.6": 361,
    "2.7": 373,
    "3.0": 361,
    "3.1": 375,
    "3.2": 392,
    "3.3": 398,
    "3.4": 429,
    "3.5": 478,
    "3.6": 494,
    "3.7": 537,
    "3.8": 569,
    "3.9": 596,
    "3.10": 619,
    "3.11": 664,
}


def _download_peps_json() -> Path:
    json_url = BASE_URL + JSON_PATH
    cache_file = _cache.filename(json_url)
    logging.info(f"Cache file: {cache_file}")

    res = _cache.load(cache_file)
    if res == {}:
        # No cache, or couldn't load cache
        import httpx

        r = httpx.get(json_url, headers={"User-Agent": USER_AGENT})

        # Raise if we made a bad request
        # (4XX client error or 5XX server error response)
        logging.info(f"HTTP status code: {r.status_code}")
        r.raise_for_status()

        res = r.json()

        _cache.save(cache_file, res)

    return cache_file


def word_search(search: str | None) -> int:
    import json

    peps_file = _download_peps_json()

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        from thefuzz import process

    with open(peps_file) as f:
        peps = json.load(f)

    # We only want to search titles
    peps = ((number, details["title"]) for number, details in peps.items())

    result: list[tuple[tuple[str, str], int]] = process.extract(search, peps)
    print("Score\tResult")
    for match, score in result:
        print(f"{score}\tPEP {match[0]}: {match[1]}")
    print()

    # Top result -> match -> PEP number
    return int(result[0][0][0])


def url(search: str | None, base_url: str = BASE_URL, pr: int | None = None) -> str:
    """Get PEP URL"""
    if pr:
        base_url = f"https://pep-previews--{pr}.org.readthedocs.build"

    result = base_url.rstrip("/")

    if search:
        try:
            # pep 8
            number = int(search)
        except ValueError:
            try:
                # pep 3.11
                number = VERSION_TO_PEP[search]
            except KeyError:
                # pep "dead batteries"
                number = word_search(search)

        result += f"/pep-{number:0>4}/"

    return result


def pep(
    search: str, base_url: str = BASE_URL, pr: int | None = None, dry_run: bool = False
) -> None:
    """Open this PEP in the browser"""
    pep_url = url(search, base_url, pr)
    if not dry_run:
        import webbrowser

        webbrowser.open_new_tab(pep_url)
    print(pep_url)



"""
Cache functions
"""
from __future__ import annotations

import datetime as dt
import json
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from platformdirs import user_cache_dir

if TYPE_CHECKING:
    if sys.version_info >= (3, 10):
        from typing import TypeAlias
    else:
        from typing_extensions import TypeAlias

PepData: TypeAlias = "dict[str, dict[str, str | None]]"

CACHE_DIR = Path(user_cache_dir("pepotron"))


def filename(url: str) -> Path:
    """yyyy-mm-dd-url-slug.json"""
    from slugify import slugify

    today = dt.datetime.utcnow().strftime("%Y-%m-%d")
    slug = slugify(url)
    return CACHE_DIR / f"{today}-{slug}.json"


def load(cache_file: Path) -> PepData:
    """Load data from cache_file"""
    if not cache_file.exists():
        return {}

    logging.info("Cache file exists")
    with cache_file.open("r") as f:
        try:
            data: PepData = json.load(f)
        except json.decoder.JSONDecodeError:
            return {}

    return data


def save(cache_file: Path, data: PepData) -> None:
    """Save data to cache_file"""
    try:
        if not CACHE_DIR.exists():
            CACHE_DIR.mkdir(parents=True)

        with cache_file.open("w") as f:
            json.dump(data, f)

    except OSError:
        pass


def clear(clear_all: bool = False) -> None:
    """Delete all or old cache files"""
    cache_files = CACHE_DIR.glob("**/*.json")
    today = dt.datetime.utcnow().strftime("%Y-%m-%d")
    for cache_file in cache_files:
        if clear_all or not cache_file.name.startswith(today):
            cache_file.unlink()



"""
pepotron: CLI to open PEPs in your browser
"""
import argparse
import atexit
import logging

from pepotron import BASE_URL, __version__, _cache, pep

atexit.register(_cache.clear)


def main() -> None:
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "search",
        nargs="*",
        help="PEP number, or Python version for its schedule, or words from title",
    )
    parser.add_argument(
        "-u", "--url", default=BASE_URL, help=f"Base URL for PEPs (default: {BASE_URL})"
    )
    parser.add_argument("-p", "--pr", type=int, help="Open preview for python/peps PR")
    parser.add_argument(
        "-n", "--dry-run", action="store_true", help="Don't open in browser"
    )
    parser.add_argument(
        "--clear-cache", action="store_true", help="Clear cache before running"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
        default=logging.WARNING,
        help="Verbose logging",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {__version__}"
    )
    args = parser.parse_args()

    logging.basicConfig(level=args.loglevel)
    if args.search:
        args.search = " ".join(args.search)
    if args.clear_cache:
        _cache.clear(clear_all=True)

    pep(search=args.search, base_url=args.url, pr=args.pr, dry_run=args.dry_run)


if __name__ == "__main__":
    main()
